package Site::SQLTable;
# LICENSE
#
# Copyright (c) 2001 Steve Slaven <bpk@hoopajoo.net> All rights reserved.
#
# This program is free software  and is provided "as is" without
# express or implied warranty. You can redistribute it and/or
# modify it under the same terms as Perl itself.
#
use strict;
use vars qw( @ISA %AUTOFIELDS $AUTOSETTINGS );

use Site::SQLContainer;
@ISA = qw( Site::SQLContainer );

use Site::Page;
use Site::Template;

$AUTOSETTINGS = 1;
%AUTOFIELDS = (
			formaction => $::ENV{ SCRIPT_NAME },
			template => undef,
			templatefile => undef,
			deletetemplate => undef,
			forwardbutton => 'Next --->',
			backbutton => '<--- Back',
			message => undef,
			pkeyvals => undef,
			whereclause => undef,
			whereargs => undef,

# Start from is **1** based
			startfrom => 1,
			endon => undef,

# 0 means all matches
			perpage => 0,
			totalrows => undef,

			expandedheaders => 1,

			query => undef,
			order => undef,
			usedata => undef,
			otherdata => undef,

			lookupobjects => 0
	);

sub OBJ_NEW
{
	my $self = shift;

	$self -> { usedata } = {
					rows => []
			};
	$self -> { otherdata } = {};

	return( $self );
}

sub display
{
	my $self = shift;
	$self -> _LOADTABLEOPTS();

	my @ToDisplay; # <--- The array of rows to display for this run... ?

	# Loop through the query.  :)
	my @pkeys = @{ $self -> pkeys() };
	my @Fields = $self -> _SELECTFIELDS();
	my @Values;

	# Generate the select query
	my $MagicQ;

	if( $self -> query() )
	{
		$MagicQ = $self -> query();
	}
	else
	{
		$MagicQ = 'select ' . join( ", ", @Fields ) .
			' from ' . $self -> table();

		# Filled in because it's used later
		$self -> query( $MagicQ );
	}

	if( $self -> whereclause() )
	{
		$MagicQ .=
			' where ' .
			$self -> whereclause();
	}

	if( $self -> order() )
	{
		$MagicQ .=  ' order by ' . join( ", ",
			@{ $self -> order() } );
	}

	my $sth = $self -> dbh() -> prepare( $MagicQ );
	if( $self -> whereargs() )
	{
		$sth -> execute( @{ $self -> whereargs() } );
	}else
	{
		$sth -> execute();
	}

	my( $ThisRow, $ThisRef, $TheField );
	my $Rows = $sth -> fetchall_arrayref({});
	my( $ThisValue, $FieldsRef, $Tmp, @MkUseRows, $start, $end );

	# Set some attribs
	$self -> totalrows( scalar( @{ $Rows } ) );
	$self -> endon( $self -> totalrows() );

	if( $self -> perpage() > 0 )
	{
		# Using startat/perpage stuff
		$start = $self -> startfrom() - 1;
		$self -> endon( $start + $self -> perpage() );
		$end = $self -> endon() - 1;

		@MkUseRows = @{ $Rows }[ $start .. $end ];
		$Rows = \@MkUseRows;

		# Add in the usedata
		$self -> usedata() -> { forward } =
			$self -> _HTMLBUTTONFROM( $self -> forwardbutton(),
						$end + 2 ); # +2, 1 to get in sync, 1 for next item

		$self -> usedata() -> { back } =
			$self -> _HTMLBUTTONFROM( $self -> backbutton(),
						$start - $self -> perpage() + 1 ); # +1 to get in sync ( off by 1 )

		$self -> usedata() -> { totalpages } = int( $self -> totalrows() / $self -> perpage() );
		if( $self -> totalrows() % $self -> perpage() )
		{
			$self -> usedata() -> { totalpages } += 1;
		}

		$self -> usedata() -> { currentpage } = int( $start / $self -> perpage() ) + 1;
	}

	$self -> usedata() -> { expandedheaders } = $self -> expandedheaders();
	$self -> usedata() -> { total } = $self -> totalrows();
	$self -> usedata() -> { startfrom } = $self -> startfrom();
	$self -> usedata() -> { endon } = $self -> endon();
	$self -> usedata() -> { width } = '100%';
	$self -> usedata() -> { perpage } = $self -> perpage();

	# Gotta put in the headers
	$self -> usedata() -> { headers } = [];
	$self -> usedata() -> { byname } = {};
	$self -> usedata() -> { byboundto } = {};
	for( @{ $self -> listfields() } )
	{
		push( @{ $self -> usedata() -> { headers } }, {
				header => $self -> _HTMLHEADER( $_ -> name() )
			} );

		$self -> usedata() -> { byname } -> { $_ -> name() } =
				$self -> _HTMLHEADER( $_ -> name() );
		$self -> usedata() -> { byboundto } -> { $_ -> boundto() } =
				$self -> _HTMLHEADER( $_ -> name() );
	}

	for( @{ $Rows } )
	{
		# Fill fields, then pull vals into ThisRow and push
		# it onto the stack of rows	
		$ThisRef = $_;
		$ThisRow = {};

		# We have to loop through the fields because
		# of the formatting, etc
		for( keys( %{ $ThisRef } ) )
		{
			$TheField = $self -> getboundfield( $_ );

			if( ! $TheField )
			{
				# Doesn't exist, too many fields?
				#carp( "Field doesn't exist: $_" );
			}else
			{
				$TheField -> _SETVALUE( $ThisRef -> { $_ } );
			}
		}

		# Also, for each unbound field, perform {} substitution on it
		for( @{ $self -> listfields() } )
		{
			# Copy/sub for non bound fields
			if( ! $_ -> boundto() )
			{
				$Tmp = $_ -> default();
				$Tmp =~ s/\{(.*?)\}/${$ThisRef}{$1}/ge;
				$_ -> _SETVALUE( $Tmp );
			}

			# Href processing
			if( $_ -> href() )
			{
				$Tmp = $_ -> href();
				$Tmp =~ s/\{(.*?)\}/${$ThisRef}{$1}/ge;

				$_ -> _SETVALUE( '<A HREF="' .
						$Tmp .
						'">' .
						$_ -> value() .
						'</A>' );
			}
		}

		# Make a snapshot of this row
		$ThisRow -> { fields } = [];
		$ThisRow -> { byname } = {};
		$ThisRow -> { byboundto } = {};

		for( @{ $self -> listfields() } )
		{
			$self -> _AddFieldToHREF( $ThisRow, $_ );
		}

		# Add the object?
		if( ( $self -> objectname() ) &&
			( $self -> lookupobjects() ) )
		{
			#print $ThisRef . ":" . $_ ;
			$Tmp = [];
			# Grab the pkey values
			for( @{ $self -> pkeys() } )
			{
				push( @{ $Tmp }, $ThisRef -> 
					{ $_ } );
			}
			$ThisRow -> { object } =
				( $self -> objectname() ) -> SQL_GET( @{ $Tmp } );
		}

		push( @{ $self -> usedata() -> { rows } }, $ThisRow );
	}

	#print $MagicQ;
	my $Ret;

	# Performs the actual form display
	if( ! $self -> template() )
	{
		$self -> template( Site::Template -> new() );

		# Check for template file...
		if( $self -> templatefile() )
		{
			$self -> template -> load( $self -> templatefile() );
		}else
		{
			$self -> template() -> template( $self -> _DEFAULTTEMPLATE() );
		}
	}

	$self -> template() -> registerdefault( $self -> usedata() );
	$self -> template() -> register( other => $self -> otherdata() );
	$self -> template() -> display();
}

sub _DEFAULTTEMPLATE
{
	# Return a default template
	return( qq{
<t:iftrue name=expandedheaders>
Showing <t:var name=startfrom/> through <t:var name=endon/> of <t:var name=total/>
<t:iftrue name=perpage><BR>Page <t:var name=currentpage/> of <t:var name=totalpages/></t:iftrue>
</t:iftrue>

<TABLE BORDER=0 WIDTH=<t:var name=width/>>
<TR>
<t:loop name=headers>
<TD><t:var name=header/></tD>
</t:loop>
</TR>

<t:loop name=rows>
<TR>
<t:loop name=fields>
<TD VALIGN=TOP><t:var name=display/></TD>
</t:loop>
</TR>
</t:loop>

</TABLE>

<t:iftrue name=perpage>
<TABLE BORDER=0 WIDTH=100\%>
<TR>
<TD><t:var name=back/></TD>
<TD ALIGN=RIGHT><t:var name=forward/></TD>
</TR>
</TABLE>
</t:iftrue>

} );

}

sub HTMLBegin
{
	my $self = shift;
	my $Action = $self -> formaction();
	my $Do = $self -> _WhatForm();

	my $Ret = qq{<FORM ACTION=$Action METHOD=GET>
} . '<INPUT TYPE=HIDDEN NAME=' . $self -> _ACTION() .
	' VALUE="' . $Do . '">';

	if( $Do ne "ADD" )
	{
		# Add pkey vals
		my @Vals = @{ $self -> pkeyvals() };
		for( @{ $self -> pkeys() } )
		{
			$Ret .= '<INPUT TYPE=HIDDEN NAME=' .
				$self -> _HTMLNAME() .
				'.pkeys.' .
				$_ .
				' VALUE="' .
				Site::Page ->
					encode_entities( shift( @Vals ) ) .
				'">';
		}
	}

	# Spew forth the carrables!
	for ( keys( %{ $self -> carry() } ) )
	{
		$Ret .= '<INPUT TYPE=HIDDEN NAME=' .
			$_ .
			' VALUE="' .
			Site::Page ->
				encode_entities( $self -> carry() -> { $_ } ) .
			'">';
	}

	return( $Ret );
}


sub HTMLEnd
{
	return( qq{</FORM>} );
}

sub htmllabel
{
	my $self = shift;
	my $Fname = shift;

	return( $self -> getnamedfield( $Fname ) -> htmllabel() );
}

sub htmlfield
{
	my $self = shift;
	my $Fname = shift;

	return( $self -> getnamedfield( $Fname ) -> htmlfield() );
}

sub HTMLAdd
{
	# The add button, if applicable
	my $self = shift;

	if($self -> _WhatForm() eq "ADD" )
	{
		return( '<INPUT TYPE=SUBMIT NAME=' . 
			$self -> _Button() .
			' VALUE="' . 
			$self -> AddButton() .
			'">' );
	}
	return( undef );
}

sub HTMLEdit
{
	# The edit button
	my $self = shift;
	if( $self -> _WhatForm() eq "UPDATE" )
	{
		return( '<INPUT TYPE=SUBMIT NAME=' . 
			$self -> _Button() .
			' VALUE="' . 
			$self -> EditButton() .
			'">' );
	}

	return( undef );
}

sub htmldisplay
{
	my $self = shift;
	return( $self -> getnamedfield( shift() ) -> displayValue() );
}

sub _Button
{
	my $self = shift;

	return( $self -> _HTMLNAME() . ".__.button" );
}

sub AddMessage
{
	my $self = shift;
	$self -> { message } .= shift() . "\n";
}

sub _HTMLPKEYS
{
	# This will be configurable later, for now the pkeys are not fully
	# qualified
	my $self = shift;

	return( $self -> pkeys() );
}

sub _HTMLHEADER
{
	my $self = shift;
	my( %NewOpts );
	my $Field = $self -> getnamedfield( shift() );
	my $SortOrder;

	if( ! $Field )
	{
		return( undef );
	}

	if( ! $Field -> boundto() )
	{
		return( $Field -> htmllabel() );
	}

	# Fill out the stuff to carry
	$NewOpts{ 'query' } = Site::Page -> escape_uri( $self -> query() );

	$NewOpts{ 'order' } = Site::Page -> escape_uri( join( ",",
					$Field -> boundto(),
					grep( $_ ne $Field -> boundto(),
						@{ $self -> order() || [] } ) ) );

	$NewOpts{ 'whereclause' } = Site::Page -> escape_uri( $self -> whereclause() );

	if( $self -> whereargs() )
	{
		$NewOpts{ 'whereargs' } = Site::Page -> escape_uri( join( "\t", @{ $self -> whereargs() } ) );
	}

	# really inefficient right now!!
	if( $self -> order() )
	{
		my $Cnt = 0;
		for( @{ $self -> order() } )
		{
			$Cnt++;
			s/\s.*$//;

			if( lc( $_ ) eq lc( $Field -> boundto() ) )
			{
				$SortOrder = "(" . $Cnt . ")";
			}
		}
	}

	return( $SortOrder . '<A HREF="' .
		$self -> formaction() .
		'?' .
		$self -> _ACTION() .
		'=SORT&' .
		join( '&', map( $_ . '=' . 
				Site::Page -> escape_uri(
					$self -> { carry } -> { $_ } ),
				keys( %{ $self -> { carry } } ) ) ) . '&' .
		join( "&", map( $self -> _HTMLNAME() . ".sort." . $_ .
					'=' . $NewOpts{ $_ }, keys( %NewOpts ) ) ) .
		'">' .
		$Field -> htmllabel() .
		'</A>' );
}

# Creates a next or last href
sub _HTMLBUTTONFROM
{
	my $self = shift;
	my( %NewOpts );

	my $Label = shift;
	my $StartAt = shift;

	my(  $SortOrder, $Link );

	# Check bounds
	if( $StartAt < 1 )
	{
		return( undef );
	}

	if( ( $StartAt + 1 ) > $self -> totalrows() )
	{
		return( undef );
	}

	# Fill out the stuff to carry
	$NewOpts{ 'startfrom' } = $StartAt;

	$NewOpts{ 'query' } = Site::Page -> escape_uri( $self -> query() );

	$NewOpts{ 'order' } = Site::Page -> escape_uri( join( ",",
						@{ $self -> order() || [] } ) );

	$NewOpts{ 'whereclause' } = Site::Page -> escape_uri( $self -> whereclause() );

	if( $self -> whereargs() )
	{
		$NewOpts{ 'whereargs' } = Site::Page -> escape_uri( join( "\t", @{ $self -> whereargs() } ) );
	}

	$Link = '<A HREF="' .
		$self -> formaction() .
		'?' .
		$self -> _ACTION() .
		'=MOVE&' .
		join( '&', map( $_ . '=' . 
				Site::Page -> escape_uri(
					$self -> { carry } -> { $_ } ),
				keys( %{ $self -> { carry } } ) ) ) . '&' .
		join( "&", map( $self -> _HTMLNAME() . ".sort." . $_ .
					'=' . $NewOpts{ $_ }, keys( %NewOpts ) ) ) .
		'">' .
		$Label .
		'</A>';

	return( $Link );
}

sub _LOADTABLEOPTS
{
	# Override table options if passed in the URI
	my $self = shift;
	my $Opts = $self -> opts();
	my $Sname = $self -> _HTMLNAME() . ".sort.";

	if( ( $Opts -> { $self -> _ACTION() } eq "SORT" ) ||
		( $Opts -> { $self -> _ACTION() } eq "MOVE" ) )
	{
		$self -> { query } = $Opts -> { $Sname . "query" };
		$self -> { whereclause } = $Opts -> { $Sname . "whereclause" };

		$self -> { order } = [];
		for( split( /,/, $Opts -> { $Sname . "order" } ) )
		{
			push( @{ $self -> { order } }, $_ );
		}

		if( ! @{ $self -> { order } } )
		{
			$self -> { order } = undef;
		}

		$self -> { whereargs } = [];
		for( split( /\t/, $Opts -> { $Sname . "whereargs" } ) )
		{
			push( @{ $self -> { whereargs } }, $_ );
		}
	}

	if( $Opts -> { $self -> _ACTION() } eq "MOVE" )
	{
		# These too
		$self -> perpage( $Opts -> { $Sname . "perpage" } );
		$self -> startfrom( $Opts -> { $Sname . "startfrom" } );
	}
}

sub _AddFieldToHREF
{
	my $self = shift;
	my $href = shift;
	my $Field = shift;

	my( $useref, $fieldref );
	
	$useref = {
			field => $Field -> htmlfield(),
			label => $Field -> htmllabel(),
			display => $Field -> htmldisplay(),
			sqlvalue => $Field -> _SQLVALUE(),
			header => $self -> _HTMLHEADER( $Field -> name() )
		};

	push( @{ $href -> { fields } }, $useref );
	$href -> { byname } -> { $Field -> name() } = $useref;
	$href -> { byboundto } -> { $Field -> boundto() } = $useref;
}

sub EXAMPLE_TEMPLATE
{
	return q{
<TABLE BORDER=1 WIDTH=100%>

<!-- Headers in the byname section.  Also could use byboundto -->
<t:with name=byname>
<TR>
<TD><t:var name=clientid/></TD>
<TD><t:var name=name/></TD>
<TD><t:var name=email/></TD>
<TD><t:var name=billemail/></TD>
</TR>
</t:with>

<!-- The rows[] loop -->
<t:loop name=rows>
<!-- Contains a fields[] for each field's display/field/header value,
     and also byname and byboundto with each fields name or boundto field
     and under that has the display/field/header values again -->
<t:with name=byname>
<TR>
<TD><t:var name=clientid:display/></TD>
<TD><t:var name=name:display/> (<t:var name=edit:display/> - <t:var name=invoices:display/>)</TD>
<TD><t:var name=email:display/></TD>
<TD><t:var name=billemail:display/></TD>
</TR>

<TR>
<TD><t:iftrue name=!..:!..:other:showrates><t:var name=rate:display/></t:iftrue></TD>
<TD><t:var name=address:display/></TD>
<TD><t:var name=city:display/></TD>
<TD><t:var name=zip:display/></TD>
</t:with>
</t:loop>

</TABLE>
};
}

sub DUMP_TEMPLATE
{
        # For making custom templates.  This dumps a template for the
        # current data
        my $self = shift();
        my( $ret, $name );

        if( ! ref( $self ) )
        {
                return( $self -> _DEFAULTTEMPLATE() );
        }

        $ret = q{
<!-- You can strip this out if you don't care for the header info -->
<t:iftrue name=expandedheaders>
Showing <t:var name=startfrom/> through <t:var name=endon/> of <t:var name=total/>
<t:iftrue name=perpage><BR>Page <t:var name=currentpage/> of <t:var name=totalpages/></t:iftrue>
</t:iftrue>

<TABLE BORDER=0 WIDTH=<t:var name=width/>>

<!-- Start with headers -->
<t:with name=byname>
<TR>
};

        for( @{ $self -> listfields() } )
        {
                $name = $_ -> name();
                $ret .= qq{
<TD><t:var name=$name/></TD>
};
        }

	$ret .= qq{
</TR>
</t:with>

<!-- Loop over rows -->
<t:loop name=rows>
<TR>
};

        for( @{ $self -> listfields() } )
        {
                $name = $_ -> name();
                $ret .= qq{
<TD><t:var name=$name:display/></TD>
};
	}

	$ret .= qq{
</TR>
</t:loop>

</TABLE>

<!-- Next/last navigation -->
<t:iftrue name=perpage>
<TABLE BORDER=0 WIDTH=100\%>
<TR>
<TD><t:var name=back/></TD>
<TD ALIGN=RIGHT><t:var name=forward/></TD>
</TR>
</TABLE>
</t:iftrue>
};

	return( $ret );
}

# where, sets where and where args
sub where
{
	my $self = shift;

	$self -> { whereclause } = shift();
	my @args = @_;
	$self -> { whereargs } = \@args;
}

sub OBJ_HOOK_startfrom
{
	my $self = shift;
	my $val = sprintf( "%01d", shift() );

	if( $val < 1 )
	{
		return( 1 );
	}
	return( $val );
}

sub OBJ_HOOK_endon
{
	my $self = shift;
	my $val = sprintf( "%01d", shift() );

	if( $val > $self -> totalrows() )
	{
		return( $self -> totalrows() );
	}
	return( $val );
}

1;

__END__

=head1 NAME

Site::SQLtable - Mostly painless SQL tables with auto-sorts,
		prev/next, etc.

=head1 SYNOPSIS

  use Site::SQLtable;

  $t = Site::SQLtable -> new( dbh => $dbh );
  $t -> table( 'customer_t' );
  $t -> addfields( { .. } );
  $t -> display();

=head1 DESCRIPTION

The SQLtable class handles the dirty underbelly of creating and using
tables with SQL data.  It can automatically do X per page lists with
column headers that allow the data to be re-sorted easily with little
programming effort.  They can also be customized with forms to change
the look and feel, see the section about templates for more information.

For basic fields information and how to get them in to an SQLtable,
see the Site::SQLContainer docs.

=head2 display

Displays the current table.

=head2 wherecluase

Sets the where part of the table, without the word 'where'.

  $f -> whereclause( 'customerid=3' );

=head2 whereargs

Used with the whereclause(), these are the args to the DBI -> execute()
call, should be an arrayref.

  $f -> whereclause( 'customerid=? AND useid=?' );
  $f -> whereargs( [ $custid, $useid ] );

=head2 where

Combines whereclause and whereargs in to 1 call:

  $f -> where( 'customerid=? AND useid=?', $custid, $useid );

is equivalent to the whereargs() example.  You don't use an arrayref here.

=head2 order

An array ref specifying the default order (when no re-sorting has been done).

  $f -> order( [ 'lastname', 'firstname' ] );

=head2 perpage

Sets the number of rows per page.  If 0, then show all rows.  Default 0.

=head2 query

Normally the query is build from the table and fields properties, but you
can override this by specifying an implicit query.  This will be used
and the field names will be bound to the correct fields.  Mostly useful
for presenting JOINed data.  Note, you MUST NOT SPECIFY A WHERE OR
ORDER BY CLAUSE!!

  $f -> query( qq{
select name, order_description, value
from customer join order using (customerid) } );

=head2 lookupobjects

If set to '1' and the table was created from a 'bind_sql_object'
then the SQLtable will create (with new() and SQL_GET()) objects
for each row to be used in a template.  It's a performance hit, so
if you don't need it then don't use it, however it makes the tables
very powerful when used, for example to have extended logic within
the table or calculated values or such.

=head2 forwardbutton/backbutton

The string to be used to create the 'NEXT'/'BACK' button link.
It is a link too, not a form button.  You could use an <IMG> if you want.

=head1 Templates

Templates work similar to SQLForm templates but there is the addition
of headers and the loop'able rows.

The default template is:

 <t:iftrue name=expandedheaders>
 Showing <t:var name=startfrom/> through
  <t:var name=endon/> of <t:var name=total/>
 <t:iftrue name=perpage><BR>Page <t:var name=currentpage/> of 
  <t:var name=totalpages/></t:iftrue>
 </t:iftrue>

 <TABLE BORDER=0 WIDTH=<t:var name=width/>>
 <TR>
 <t:loop name=headers>
  <TD><t:var name=header/></tD>
 </t:loop>
 </TR>

 <t:loop name=rows>
 <TR>
 <t:loop name=fields>
 <TD VALIGN=TOP><t:var name=display/></TD>
 </t:loop>
 </TR>
 </t:loop>

 </TABLE>

 <t:iftrue name=perpage>
 <TABLE BORDER=0 WIDTH=100\%>
 <TR>
 <TD><t:var name=back/></TD>
 <TD ALIGN=RIGHT><t:var name=forward/></TD>
 </TR>
 </TABLE>
 </t:iftrue>

This can be customized by loading an external template.

=head2 expandedheaders

Just passed from the class, you can probably ignore it since you know
if you need expanded headers or not.  It was in the default template
to allow somewhat easier settings from the class itself.

=head2 startfrom/endon/total/perpage/currentpage

All just informational things to be displayed to the user if you want.

=head2 headers

A loop'able variable containing all the headers, which is the re-sort
links for the table.  Under the headers is just a header variable.

=head2 byname/byboundto

At the very top level these are the HEADERS by name and boundto fields.
Under the var is just the names, which are the headers for the fields.

=head2 rows

The rows to be displayed, under the rows you have several options:

=head2 object

If lookupobjects is '1' then this will contain an SQLObject for the
current row.

=head2 fields

A loop'able var containing all the fields.

=head2 byname/byboundto

Fields by name and boundto field name.

=head2 THE FIELDS

The fields all contain a field, label, display, sqlvalue and header
variable.  That is for both fields and the boundto/name lists.

=head1 AUTHOR

Copyright (c) 2001 Steve Slaven <bpk@hoopajoo.net> All rights reserved.

This program is free software  and is provided "as is" without
express or implied warranty. You can redistribute it and/or
modify it under the same terms as Perl itself.

=head1 SEE ALSO

perl(1).

=cut
